import itertools
import bpy
import mathutils

class MeshExporter:
	def __init__(self):
		self.show_progress = True
		self.use_strips = True
		self.use_degen_tris = False
		self.material_tex = False

	def join_strips(self, strips):
		big_strip = []

		for s in strips:
			if big_strip:
				# Generate glue elements, ensuring that the next strip begins at
				# an even position
				glue = [big_strip[-1], s[0]]
				if len(big_strip)%2:
					glue += [s[0]]

				big_strip += glue

			big_strip += s

		return big_strip

	def export(self, context, out_file, obj=None, progress=None):
		if obj is None:
			obj = context.active_object

		from .mesh import create_mesh_from_object
		from .util import Progress

		if not progress:
			progress = Progress(self.show_progress and context)
		progress.push_task("", 0.0, 0.9)

		mesh = create_mesh_from_object(context, obj, progress)

		strips = []
		loose = mesh.faces
		if self.use_strips:
			strips = mesh.vertex_sequence
			if self.use_degen_tris:
				strips = [self.join_strips(strips)]
			loose = []

		progress.set_task("Writing file", 0.9, 1.0)

		from .outfile import open_output
		out_file = open_output(out_file)

		fmt = ["NORMAL3"]
		if mesh.uv_layers:
			for u in mesh.uv_layers:
				size = str(len(u.uvs[0]))
				if u.unit==0:
					fmt.append("TEXCOORD"+size)
				else:
					fmt.append("TEXCOORD%s_%d"%(size, u.unit))
			if mesh.tbn_vecs:
				fmt += ["TANGENT3", "BINORMAL3"]
		if mesh.vertex_groups:
			fmt.append("ATTRIB%d_5"%(mesh.max_groups_per_vertex*2))
		fmt.append("VERTEX3")
		out_file.begin("vertices", *fmt)
		normal = None
		uvs = {}
		tan = None
		bino = None
		group = None
		for v in mesh.vertices:
			if v.normal!=normal:
				out_file.write("normal3", *v.normal)
				normal = v.normal
			for i, u in enumerate(mesh.uv_layers):
				if v.uvs[i]!=uvs.get(i):
					size = str(len(v.uvs[i]))
					if u.unit==0:
						out_file.write("texcoord"+size, *v.uvs[i])
					else:
						out_file.write("multitexcoord"+size, u.unit, *v.uvs[i])
					uvs[i] = v.uvs[i]
			if mesh.tbn_vecs:
				if v.tan!=tan:
					out_file.write("tangent3", *v.tan)
					tan = v.tan
				if v.bino!=bino:
					out_file.write("binormal3", *v.bino)
					bino = v.bino
			if mesh.vertex_groups:
				group_attr = [(group_index_map[g.group], g.weight*v.group_weight_scale) for g in v.groups[:mesh.max_groups_per_vertex]]
				while len(group_attr)<mesh.max_groups_per_vertex:
					group_attr.append((0, 0.0))
				group_attr = list(itertools.chain(*group_attr))
				if group_attr!=group:
					out_file.write("attrib%d"%len(group_attr), 5, *group_attr)
					group = group_attr
			out_file.write("vertex3", *v.co)
		out_file.end()
		for s in strips:
			out_file.begin("batch", "TRIANGLE_STRIP")
			indices = []
			n = 0
			for v in s:
				indices.append(v.index)
				if len(indices)>=32:
					out_file.write("indices", *indices)
					indices = []
			if indices:
				out_file.write("indices", *indices)
			out_file.end()

		if loose:
			out_file.begin("batch", "TRIANGLES")
			for f in loose:
				for i in range(2, len(f.vertices)):
					out_file.write("indices", f.vertices[0].index, f.vertices[i-1].index, f.vertices[i].index)
			out_file.end()

		if mesh.lines:
			out_file.begin("batch", "LINES")
			for l in mesh.lines:
				out_file.write("indices", l.vertices[0].index, l.vertices[1].index)
			out_file.end()

		if mesh.winding_test:
			out_file.write("winding", "COUNTERCLOCKWISE")

		progress.pop_task()

		return mesh
