import bpy
import math
import mathutils

def make_edge_key(i1, i2):
	return (min(i1, i2), max(i1, i2))

class Edge:
	def __init__(self, edge):
		if edge.__class__==Edge:
			self._edge = edge._edge
			self.smooth = edge.smooth
		else:
			self._edge = edge
			self.smooth = False
		if edge:
			self.vertices = edge.vertices[:]
			self.key = edge.key
		else:
			self.vertices = []
			self.key = None
		self.faces = []

	def __getattr__(self, attr):
		return getattr(self._edge, attr)

	def check_smooth(self, limit):
		if len(self.faces)!=2:
			return

		d = self.faces[0].normal.dot(self.faces[1].normal)
		self.smooth = ((d>limit and self.faces[0].use_smooth and self.faces[1].use_smooth) or d>0.99995)

	def other_face(self, f):
		if f.index==self.faces[0].index:
			if len(self.faces)>=2:
				return self.faces[1]
			else:
				return None
		else:
			return self.faces[0]

	def other_vertex(self, v):
		if v.index==self.vertices[0].index:
			return self.vertices[1]
		else:
			return self.vertices[0]


class Vertex:
	def __init__(self, vertex):
		if vertex.__class__==Vertex:
			self._vertex = vertex._vertex
			self.uvs = vertex.uvs[:]
			self.tan = vertex.tan
			self.bino = vertex.bino
		else:
			self._vertex = vertex
			self.uvs = []
			self.tan = None
			self.bino = None
		self.index = vertex.index
		self.co = vertex.co
		self.normal = vertex.normal
		self.flag = False
		self.edges = []
		self.faces = []
		self.groups = vertex.groups[:]

	def __getattr__(self, attr):
		return getattr(self._vertex, attr)

	def __cmp__(self, other):
		if other is None:
			return 1
		return cmp(self.index, other.index)


class VertexGroup:
	def __init__(self, group):
		self._group = group
		self.group = group.group
		self.weight = group.weight

	def __getattr__(self, attr):
		return getattr(self._group, attr)


class Face:
	def __init__(self, face):
		self._face = face
		self.index = face.index
		self.edges = []
		self.vertices = face.vertices[:]
		self.uvs = []
		self.flag = False

	def __getattr__(self, attr):
		return getattr(self._face, attr)

	def __cmp__(self, other):
		if other is None:
			return 1
		return cmp(self.index, other.index)

	def pivot_vertex(self, v):
		n = self.vertices.index(v)
		return [(n+i)%len(self.vertices) for i in range(len(self.vertices))]

	def pivot_vertices(self, *vt):
		flags = [(v in vt) for v in self.vertices]
		l = len(self.vertices)
		for i in range(l):
			if flags[i] and not flags[(i+l-1)%l]:
				return self.vertices[i:]+self.vertices[:i]

	def get_edge(self, v1, v2):	
		key = make_edge_key(v1.index, v2.index)
		for e in self.edges:
			if e.key==key:
				return e
		raise KeyError("No edge %s"%(key,))

	def other_edge(self, e, v):
		for d in self.edges:
			if d!=e and v in d.vertices:
				return d

	def get_neighbors(self):
		neighbors = [e.other_face(self) for e in self.edges]
		return list(filter(bool, neighbors))


class Line:
	def __init__(self, e):
		self.edge = e
		self.vertices = e.vertices[:]
		self.flag = False


class UvLayer:
	def __init__(self, arg):
		if type(arg)==str:
			self._layer = None
			self.name = arg
		else:
			self._layer = arg
			self.name = arg.name
			self.uvs = [d.uv for d in self.data]

		self.unit = None
		self.hidden = False

		dot = self.name.find('.')
		if dot>=0:
			ext = self.name[dot:]
			if ext.startswith(".unit") and ext[5:].isdigit():
				self.unit = int(ext[5:])
			elif ext==".hidden":
				self.hidden = True

	def __getattr__(self, attr):
		return getattr(self._layer, attr)


class Mesh:
	def __init__(self, mesh):
		self._mesh = mesh

		self.winding_test = mesh.winding_test
		self.tbn_vecs = mesh.tbn_vecs
		self.vertex_groups = mesh.vertex_groups

		# Clone basic data
		self.vertices = [Vertex(v) for v in self.vertices]
		for v in self.vertices:
			v.groups = [VertexGroup(g) for g in v.groups]

		self.faces = [Face(f) for f in self.polygons]
		self.edges = [Edge(e) for e in self.edges]
		self.loops = self.loops[:]
		self.materials = self.materials[:]

		# Clone only the desired UV layers
		if self.use_uv=='NONE' or not self.uv_layers:
			self.uv_layers = []
		else:
			self.uv_layers = [UvLayer(u) for u in self.uv_layers]
			self.uv_layers = sorted([u for u in self.uv_layers if not u.hidden], key=(lambda u: (u.unit or 1000, u.name)))

			if self.use_uv=='UNIT0':
				self.uv_layers = [self.uv_layers[0]]

			# Assign texture unit numbers to UV layers that lack one
			next_unit = max((u.unit+1 for u in self.uv_layers if u.unit is not None), default=0)
			for u in self.uv_layers:
				if not u.unit:
					u.unit = next_unit
					next_unit += 1

		# Rewrite links between elements to point to cloned data, or create links
		# where they don't exist
		edge_map = {e.key: e for e in self.edges}
		for f in self.faces:
			if len(f.vertices)>4:
				raise ValueError("Ngons are not supported")

			f.vertices = [self.vertices[i] for i in f.vertices]
			for v in f.vertices:
				v.faces.append(f)

			for k in f.edge_keys:
				e = edge_map[k]
				e.faces.append(f)
				f.edges.append(e)

		for e in self.edges:
			e.vertices = [self.vertices[i] for i in e.vertices]
			for v in e.vertices:
				v.edges.append(e)

		# Store loose edges as lines
		if self.use_lines:
			self.lines = [Line(e) for e in self.edges if not e.faces]
		else:
			self.lines = []

		self.vertex_sequence = []

	def __getattr__(self, attr):
		return getattr(self._mesh, attr)

	def transform(self, matrix):
		for v in self.vertices:
			v.co = matrix*v.co

	def splice(self, other):
		if len(self.uv_layers)!=len(other.uv_layers):
			raise ValueError("Meshes have incompatible UV layers")
		for i, u in enumerate(self.uv_layers):
			if u.name!=other.uv_layers[i].name:
				raise ValueError("Meshes have incompatible UV layers")

		# Merge materials and form a lookup from source material indices to the
		# merged material list
		material_map = []
		for m in other.materials:
			if m in self.materials:
				material_map.append(self.materials.index(m))
			else:
				material_map.append(len(self.materials))
				self.materials.append(m)

		# Append data and adjust indices where necessary.  Since the data is
		# spliced from the source mesh, rebuilding references is not necessary.
		for i, u in enumerate(self.uv_layers):
			u.uvs += other.uv_layers[i].uvs

		offset = len(self.vertices)
		self.vertices += other.vertices
		for v in self.vertices[offset:]:
			v.index += offset

		loop_offset = len(self.loops)
		self.loops += other.loops

		offset = len(self.faces)
		self.faces += other.faces
		for f in self.faces[offset:]:
			f.index += offset
			f.loop_start += loop_offset
			f.loop_indices = range(f.loop_start, f.loop_start+f.loop_total)
			if other.materials:
				f.material_index = material_map[f.material_index]

		offset = len(self.edges)
		self.edges += other.edges
		for e in self.edges[offset:]:
			e.index += offset
			e.key = make_edge_key(e.vertices[0].index, e.vertices[1].index)

		self.lines += other.lines

	def prepare_triangles(self, progress):
		face_count = len(self.faces)
		for i in range(face_count):
			f = self.faces[i]
			nverts = len(f.vertices)
			if nverts==3:
				continue

			# Calculate normals at each vertex of the face
			edge_vecs = []
			for j in range(nverts):
				edge_vecs.append(f.vertices[(j+1)%nverts].co-f.vertices[j].co)

			normals = []
			for j in range(nverts):
				normals.append(edge_vecs[j-1].cross(edge_vecs[j]).normalized())

			# Check which diagonal results in a flatter triangulation
			flatness1 = normals[0].dot(normals[2])
			flatness2 = normals[1].dot(normals[3])
			cut_index = 1 if flatness1>flatness2 else 0

			nf = Face(f)
			nf.index = len(self.faces)
			self.faces.append(nf)

			ne = Edge(None)
			ne.index = len(self.edges)
			self.edges.append(ne)

			nf.vertices = [f.vertices[cut_index], f.vertices[2], f.vertices[3]]
			nf.loop_indices = [f.loop_indices[cut_index], f.loop_indices[2], f.loop_indices[3]]
			for v in nf.vertices:
				v.faces.append(nf)

			ne.vertices = [f.vertices[cut_index], f.vertices[2+cut_index]]
			for v in ne.vertices:
				v.edges.append(ne)
			ne.key = make_edge_key(ne.vertices[0].index, ne.vertices[1].index)
			ne.smooth = True

			f.vertices[3-cut_index].faces.remove(f)
			del f.vertices[3-cut_index]
			f.loop_indices = [f.loop_indices[0], f.loop_indices[1], f.loop_indices[2+cut_index]]

			ne.faces = [f, nf]
			if cut_index==0:
				nf.edges = [ne, f.edges[2], f.edges[3]]
				f.edges = [f.edges[0], f.edges[1], ne]
			else:
				nf.edges = [f.edges[1], f.edges[2], ne]
				f.edges = [f.edges[0], ne, f.edges[3]]
			for e in nf.edges:
				if e!=ne:
					e.faces.remove(f)
					e.faces.append(nf)

			f.normal = normals[1-cut_index]
			nf.normal = normals[3-cut_index]

			progress.set_progress(i/face_count)

	def prepare_smoothing(self, progress):
		smooth_limit = -1
		if self.smoothing=='NONE':
			for f in self.faces:
				f.use_smooth = False

			smooth_limit = 1
		elif self.use_auto_smooth:
			smooth_limit = math.cos(self.auto_smooth_angle)

		for e in self.edges:
			e.check_smooth(smooth_limit)

		progress.push_task("Sharp edges", 0.0, 0.7)
		self.split_vertices(self.find_smooth_group, progress)

		if self.smoothing!='BLENDER':
			progress.set_task("Updating normals", 0.7, 1.0)
			self.compute_normals(progress)

		progress.pop_task()

	def prepare_vertex_groups(self, obj):
		for v in self.vertices:
			if v.groups:
				weight_sum = sum(g.weight for g in v.groups)
				v.groups = sorted(v.groups, key=(lambda g: g.weight), reverse=True)[:self.max_groups_per_vertex]
				weight_scale = weight_sum/sum(g.weight for g in v.groups)
				for g in v.groups:
					g.weight *= weight_scale

		if obj.parent and obj.parent.type=="ARMATURE":
			armature = obj.parent.data
			bone_indices = {b.name: i for i, b in enumerate(armature.bones)}
			group_index_map = {i: i for i in range(len(obj.vertex_groups))}
			for g in first_obj.vertex_groups:
				if g.name in bone_indices:
					group_index_map[g.index] = bone_indices[g.name]

			for v in self.vertices:
				for g in v.groups:
					g.group = group_index_map[g.group]

	def prepare_uv(self, obj, progress):
		if obj.material_tex and self.use_uv!='NONE':
			layer = UvLayer("material_tex")

			if self.use_uv=='UNIT0':
				self.uv_layers = [layer]
				layer.unit = 0
			else:
				self.uv_layers.append(layer)
				layer.unit = max((u.unit+1 for u in self.uv_layers if u.unit is not None), default=0)

			layer.uvs = [None]*len(self.loops)
			for f in self.faces:
				uv = mathutils.Vector(((f.material_index+0.5)/len(self.materials), 0.5))
				for i in f.loop_indices:
					layer.uvs[i] = uv

		# Form a list of UV layers referenced by materials with the array atlas
		# property set
		array_uv_layers = [t.uv_layer for m in self.materials if m.array_atlas for t in m.texture_slots if t and t.texture_coords=='UV']
		array_uv_layers = [u for u in self.uv_layers if u.name in array_uv_layers]

		if array_uv_layers:
			for f in self.faces:
				layer = 0
				if f.material_index<len(self.materials):
					mat = self.materials[f.material_index]
					if mat and mat.array_atlas:
						layer = mat.array_layer

				for l in array_uv_layers:
					for i in f.loop_indices:
						l.uvs[i] = mathutils.Vector((*l.uvs[i], layer))

		# Copy UVs from layers to faces
		for f in self.faces:
			for u in self.uv_layers:
				f.uvs.append([u.uvs[i] for i in f.loop_indices])

		prog_count = len(self.uv_layers)
		prog_step = 0

		# Split by the UV layer used for TBN vectors first so connectivity
		# remains intact for TBN vector computation
		tbn_layer_index = -1
		if self.tbn_vecs:
			uv_names = [u.name for u in self.uv_layers]
			if self.tbn_uvtex in uv_names:
				prog_count += 1
				tbn_layer_index = uv_names.index(self.tbn_uvtex)
				progress.push_task_slice("Computing TBN", 0, prog_count)
				self.split_vertices(self.find_uv_group, progress, tbn_layer_index)
				progress.set_task_slice(self.tbn_uvtex, 1, prog_count)
				self.compute_tbn(tbn_layer_index, progress)
				progress.pop_task()
				prog_step = 2

		# Split by the remaining UV layers
		for i, u in enumerate(self.uv_layers):
			if i==tbn_layer_index:
				continue

			progress.push_task_slice(u.name, prog_step, prog_count)
			self.split_vertices(self.find_uv_group, progress, i)
			progress.pop_task()
			prog_step += 1

		# Copy UVs from faces to vertices
		for v in self.vertices:
			if v.faces:
				# All faces still connected to the vertex have the same UV value
				f = v.faces[0]
				i = f.vertices.index(v)
				v.uvs = [u[i] for u in f.uvs]
			else:
				v.uvs = [(0.0, 0.0)]*len(self.uv_layers)

	def split_vertices(self, find_group_func, progress, *args):
		vertex_count = len(self.vertices)
		for i in range(vertex_count):
			v = self.vertices[i]
			for f in v.faces:
				f.flag = False

			# Find all groups of faces on this vertex
			groups = []
			for f in v.faces:
				if not f.flag:
					groups.append(find_group_func(v, f, *args))

			# Give groups after the first separate copies of the vertex
			for g in groups[1:]:
				nv = Vertex(v)
				nv.index = len(self.vertices)
				self.vertices.append(nv)

				for e in v.edges:
					e_faces_in_g = [f for f in e.faces if f in g]
					if not e_faces_in_g:
						continue

					if len(e_faces_in_g)<len(e.faces):
						# Create a copy of an edge at the boundary of the group
						ne = Edge(e)
						ne.index = len(self.edges)
						self.edges.append(ne)

						ne.other_vertex(v).edges.append(ne)

						for f in e_faces_in_g:
							e.faces.remove(f)
							f.edges[f.edges.index(e)] = ne
							ne.faces.append(f)

						e = ne

					e.vertices[e.vertices.index(v)] = nv
					nv.edges.append(e)

					e.key = make_edge_key(e.vertices[0].index, e.vertices[1].index)

				# Filter out any edges that were removed from the original vertex
				v.edges = [e for e in v.edges if v in e.vertices]

				for f in g:
					v.faces.remove(f)
					f.vertices[f.vertices.index(v)] = nv
					nv.faces.append(f)

			progress.set_progress(i/vertex_count)

	def find_smooth_group(self, vertex, face):
		face.flag = True

		edges = [e for e in face.edges if vertex in e.vertices]

		group = [face]
		for e in edges:
			f = face
			while e.smooth:
				f = e.other_face(f)
				if not f or f.flag:
					break

				f.flag = True
				group.append(f)
				e = f.other_edge(e, vertex)

		return group

	def find_uv_group(self, vertex, face, index):
		uv = face.uvs[index][face.vertices.index(vertex)]
		face.flag = True

		group = [face]
		for f in vertex.faces:
			if not f.flag and f.uvs[index][f.vertices.index(vertex)]==uv:
				f.flag = True
				group.append(f)

		return group

	def compute_normals(self, progress):
		for i, v in enumerate(self.vertices):
			v.normal = mathutils.Vector()
			for f in v.faces:
				fv = f.pivot_vertices(v)
				edge1 = fv[1].co-fv[0].co
				edge2 = fv[-1].co-fv[0].co
				if edge1.length and edge2.length:
					# Use the angle between edges as a weighting factor.  This gives
					# more consistent normals on bends with an inequal number of
					# faces on each side.
					v.normal += f.normal*edge1.angle(edge2)

			if v.normal.length:
				v.normal.normalize()
			else:
				v.normal = mathutils.Vector((0, 0, 1))

			progress.set_progress(i/len(self.vertices))

	def compute_tbn(self, index, progress):
		# This function is called at an early stage during UV preparation when
		# face UVs are not available yet
		layer_uvs = self.uv_layers[index].uvs

		for i, v in enumerate(self.vertices):
			v.tan = mathutils.Vector()
			v.bino = mathutils.Vector()
			for f in v.faces:
				vi = f.pivot_vertex(v)
				uv0 = layer_uvs[f.loop_indices[vi[0]]]
				uv1 = layer_uvs[f.loop_indices[vi[1]]]
				uv2 = layer_uvs[f.loop_indices[vi[-1]]]
				du1 = uv1[0]-uv0[0]
				du2 = uv2[0]-uv0[0]
				dv1 = uv1[1]-uv0[1]
				dv2 = uv2[1]-uv0[1]
				edge1 = f.vertices[vi[1]].co-f.vertices[vi[0]].co
				edge2 = f.vertices[vi[-1]].co-f.vertices[vi[0]].co
				div = (du1*dv2-du2*dv1)
				if div:
					mul = edge1.angle(edge2)/div
					v.tan += (edge1*dv2-edge2*dv1)*mul
					v.bino += (edge2*du1-edge1*du2)*mul

			if v.tan.length:
				v.tan.normalize()
			if v.bino.length:
				v.bino.normalize()

			progress.set_progress(i/len(self.vertices))

	def prepare_sequence(self, progress):
		progress.push_task("Reordering faces", 0.0, 0.5)
		self.reorder_faces(progress)

		progress.set_task("Building sequence", 0.5, 1.0)
		sequence = None
		for i, f in enumerate(self.faces):
			if sequence:
				if len(sequence)==3:
					# Rotate the first three vertices so that the new face can be added
					if sequence[0] in f.vertices and sequence[1] not in f.vertices:
						sequence.append(sequence[0])
						del sequence[0]
					elif sequence[2] not in f.vertices and sequence[1] in f.vertices:
						sequence.insert(0, sequence[-1])
						del sequence[-1]

				if sequence[-1] not in f.vertices:
					sequence = None
				else:
					to_add = [v for v in f.vertices if v!=sequence[-1] and v!=sequence[-2]]
					if len(to_add)==2:
						if (f.vertices[1]==sequence[-1]) != (len(sequence)%2==1):
							to_add.reverse()
						sequence.append(sequence[-1])
					sequence += to_add

			if not sequence:
				sequence = f.vertices[:]
				self.vertex_sequence.append(sequence)

			progress.set_progress(i/len(self.faces))

		progress.pop_task()

		self.reorder_vertices()

	def reorder_faces(self, progress):
		# Tom Forsyth's vertex cache optimization algorithm
		# http://eelpi.gotdns.org/papers/fast_vert_cache_opt.html

		for f in self.faces:
			f.flag = False

		last_triangle_score = 0.75
		cache_decay_power = 1.5
		valence_boost_scale = 2.0
		valence_boost_power = -0.5

		max_cache_size = 32
		cached_vertices = []

		# Keep track of the score and number of unused faces for each vertex
		vertex_info = [[0, len(v.faces)] for v in self.vertices]
		for vi in vertex_info:
			vi[0] = valence_boost_scale*(vi[1]**valence_boost_power)

		face = None
		reordered_faces = []

		n_processed = 0
		while 1:
			if not face:
				# Previous iteration gave no candidate for best face (or this is
				# the first iteration).  Scan all faces for the highest score.
				best_score = 0
				for f in self.faces:
					if f.flag:
						continue

					score = sum(vertex_info[v.index][0] for v in f.vertices)
					if score>best_score:
						best_score = score
						face = f

			if not face:
				break

			reordered_faces.append(face)
			face.flag = True

			for v in face.vertices:
				vertex_info[v.index][1] -= 1

				# Shuffle the vertex into the front of the cache
				if v in cached_vertices:
					cached_vertices.remove(v)
				cached_vertices.insert(0, v)

			# Update scores for all vertices in the cache
			for i, v in enumerate(cached_vertices):
				score = 0
				if i<3:
					score += last_triangle_score
				elif i<max_cache_size:
					score += (1-(i-3)/(max_cache_size-3))**cache_decay_power
				if vertex_info[v.index][1]:
					score += valence_boost_scale*(vertex_info[v.index][1]**valence_boost_power)
				vertex_info[v.index][0] = score

			face = None
			best_score = 0
			for v in cached_vertices:
				for f in v.faces:
					if not f.flag:
						score = sum(vertex_info[fv.index][0] for fv in f.vertices)
						if score>best_score:
							best_score = score
							face = f

			del cached_vertices[max_cache_size:]

			n_processed += 1
			progress.set_progress(n_processed/len(self.faces))

		self.faces = reordered_faces
		for i, f in enumerate(self.faces):
			f.index = i

	def reorder_vertices(self):
		for v in self.vertices:
			v.index = -1

		reordered_vertices = []
		for s in self.vertex_sequence:
			for v in s:
				if v.index<0:
					v.index = len(reordered_vertices)
					reordered_vertices.append(v)

		self.vertices = reordered_vertices

		for e in self.edges:
			e.key = make_edge_key(e.vertices[0].index, e.vertices[1].index)

	def drop_references(self):
		for v in self.vertices:
			v._vertex = None
			for g in v.groups:
				g._group = None
		for e in self.edges:
			e._edge = None
		for f in self.faces:
			f._face = None
		for u in self.uv_layers:
			u._layer = None
		self._mesh = None


def create_mesh_from_object(context, obj, progress):
	if obj.type!="MESH":
		raise Exception("Object is not a mesh")

	progress.push_task("Preparing mesh", 0.0, 0.2)

	objs = [(obj, mathutils.Matrix())]
	i = 0
	while i<len(objs):
		o, m = objs[i]
		i += 1
		for c in o.children:
			if c.type=="MESH" and c.compound:
				objs.append((c, m*c.matrix_local))

	mesh = None
	bmeshes = []
	for o, m in objs:
		bmesh = o.to_mesh(context.scene, True, "PREVIEW")
		bmeshes.append(bmesh)

		# Object.to_mesh does not copy custom properties
		bmesh.winding_test = o.data.winding_test
		bmesh.smoothing = o.data.smoothing
		bmesh.use_lines = o.data.use_lines
		bmesh.vertex_groups = o.data.vertex_groups
		bmesh.max_groups_per_vertex = o.data.max_groups_per_vertex
		bmesh.use_uv = o.data.use_uv
		bmesh.tbn_vecs = o.data.tbn_vecs
		bmesh.tbn_uvtex = o.data.tbn_uvtex

		me = Mesh(bmesh)
		me.transform(m)

		if mesh:
			mesh.splice(me)
		else:
			mesh = me

	progress.set_task("Triangulating", 0.2, 0.3)
	mesh.prepare_triangles(progress)
	progress.set_task("Smoothing", 0.3, 0.5)
	mesh.prepare_smoothing(progress)
	progress.set_task("Vertex groups", 0.5, 0.6)
	mesh.prepare_vertex_groups(obj)
	progress.set_task("Preparing UVs", 0.6, 0.8)
	mesh.prepare_uv(obj, progress)
	progress.set_task("Render sequence", 0.8, 1.0)
	mesh.prepare_sequence(progress)

	# Discard the temporary Blender meshes after making sure there's no
	# references to the data
	mesh.drop_references()
	for m in bmeshes:
		bpy.data.meshes.remove(m)

	progress.pop_task()

	return mesh
